-- Programmiersprachen SS 2008
-- Alex Ahammer, Alex Fischer, Klaus Purer
-- Beispiel 2: Softwareverwaltung 

module Main (main)
where

--Import
import Directory (doesFileExist)
import System (exitWith,ExitCode(ExitSuccess))
import Data.List (isInfixOf)
import Data.Char (isDigit, digitToInt, toLower)


--Konstanten
savefile = "sw.data"

type Titel = String
type AutorInnen = String
type Entstehungszeit = String
type Kommentar = String
type Versnr = Int
type Typ = String
type Adresse= String
type Person = String
type DatentraegerId = Int

data Software = Software Titel AutorInnen Entstehungszeit Kommentar deriving (Show, Read, Eq)

data Version = Version Versnr Software deriving (Show, Read)

data Datentraeger = PhysDatentraeger DatentraegerId Typ [Version] Person | 
		    AbstrDatentraeger DatentraegerId Typ [Version] Adresse deriving (Show, Read)

data Ausleihe = Ausleihe DatentraegerId Person deriving (Show, Read)

type Database = ([Software],[Version],[Datentraeger],[Ausleihe])

-- Hauptprogramm
-- Startpunkt der Softwareverwaltung
main :: IO()
main =  do  show_help
            db <- load
            cmdLoop db

         
-- Endlosschleife zur Kommandoasuführung
cmdLoop :: Database -> IO()
cmdLoop (sw,vs,dt,al) = do
	putStr "Kommando:>> "
	cmd <- getLine
	-- zuerst die IO Funktionen
	case (command cmd) of
	    "s" -> search sw dt (cmd_rest cmd)
	    "h" -> show_help
	    "q" -> exit_save db
	    "v" -> dat_test dt (parseID (arg1 cmd))
	    "a" -> dat_test dt (parseID (arg1 cmd))
	    "t" -> putStrLn "Testdaten werden geladen."
	    _   -> putStrLn "Error: Unbekanntes Kommando!"
	-- jetzt DB-Änderungen
	let db_new = case (command cmd) of
		"v" -> (sw,vs,(sell dt (parseID (arg1 cmd)) (arg2 cmd)),al)
		"a" -> (sw,vs,dt,(borrow al dt (parseID (arg1 cmd)) (arg2 cmd)))
		"t" -> test_db
		_ -> db
	--Neustarten der Schleife
	cmdLoop db_new
	where db = (sw,vs,dt,al)
	
-- Gibt den Namen eines Kommandos zurück
command :: String -> String
command cmd = (words cmd) !! 0

-- Gibt das erste Argument eines Kommandos zurück
arg1 :: String -> String
arg1 cmd = (words cmd) !! 1

-- Gibt das zweite Argument eines Kommandos zurück (bis zum Kommandoende)
arg2 :: String -> String
arg2 cmd = cmd_rest (cmd_rest cmd)

-- Gibt die Argumente eines Kommandos zurück
cmd_rest :: String -> String
cmd_rest cmd = drop ((length (command cmd))+1) cmd


-- Datenbank aus der Datei "sw.data" im selben Verzeichnis laden.
-- Wenn die Datei nicht existiert, dann leere Datenbank verwenden.
load :: IO Database
load = do   ex <- doesFileExist savefile
    	    if ex then do d <- readFile savefile
		          return (read d)
    		      else do return ([],[],[],[])

      
-- Daten speichern und Programm beenden
exit_save :: Database -> IO()
exit_save db = do   writeFile savefile (show db)
                    exitWith ExitSuccess
                    
-- Hilfe sowie Kommandos und Synopsis anzeigen
show_help :: IO()
show_help = do  putStrLn "                 **********************"
                putStrLn "                 * Softwareverwaltung *"
                putStrLn "                 **********************"
                putStrLn ""
                putStrLn "Testdaten laden:              t"
                putStrLn "Software suchen:              s <Suchbegriff>"
                putStrLn "Datenträger verkaufen:        v <Datenträger ID> <neue EigentümerIn>"
                putStrLn "Datenträger ausborgen an:     a <Datenträger ID> <Person>"
                putStrLn "Speichern und Beenden:        q"
                putStrLn "Hilfe bzw. Kommandoübersicht: h"
                

-- Software Liste nach einem Suchwort durchsuchen
search :: [Software] -> [Datentraeger] -> String -> IO()
search [] _ _            = putStr ""
search (sw:sws) dt string = do search_sw sw dt string
                               search sws dt string
                                
-- Attribute einer Software nach dem Suchwort durchsuchen, case insensitiv
-- Dazugehörige Datentraeger werden ebenfalls gelistet
search_sw :: Software -> [Datentraeger] -> String -> IO()
search_sw (Software t a e k) dt string = if (isInfixOf (map toLower string) (map toLower t) ||
                                            isInfixOf (map toLower string) (map toLower a) ||
                                            isInfixOf (map toLower string) (map toLower e) ||
                                            isInfixOf (map toLower string) (map toLower k)) 
                                            then do print_sw (Software t a e k)
                                                    find_dat (Software t a e k) dt
                                          else do putStr ""

-- Gibt eine Software formatiert aus
print_sw :: Software -> IO()
print_sw (Software t a e k) = do putStrLn "Software"
                                 putStrLn (" Titel: " ++ t)
                                 putStrLn (" AutorInnen: " ++ a)
                                 putStrLn (" Entstehungszeit: " ++ e)
                                 putStrLn (" Kommentar: " ++ k)

-- Finde Datenträger, auf denen diese Software liegt
find_dat :: Software -> [Datentraeger] -> IO()
find_dat _ []                                  = putStrLn ""
find_dat sw ((PhysDatentraeger id t v p):dts)  = do find_vers sw (PhysDatentraeger id t v p) v
                                                    find_dat sw dts
find_dat sw ((AbstrDatentraeger id t v a):dts) = do find_vers sw (AbstrDatentraeger id t v a) v
                                                    find_dat sw dts

-- Durchsuche die Versionen auf einem Datenträger nach der gewünschten Software
find_vers :: Software -> Datentraeger -> [Version] -> IO()
find_vers sw dt [] = putStr ""
find_vers sw1 dt ((Version nr sw2):vs)    
    | sw1 == sw2  = do print_dat dt (Version nr sw2)
                       find_vers sw1 dt vs
    | otherwise   = find_vers sw1 dt vs

-- Gibt einen Datenträger formatiert aus
print_dat :: Datentraeger -> Version -> IO()
print_dat (PhysDatentraeger id t v p) (Version nr sw) = 
    do putStrLn "Physischer Datenträger"
       putStrLn (" ID: " ++ (show id))
       putStrLn (" Typ: " ++ t)
       putStrLn (" Version: " ++ (show nr))
       putStrLn (" EigentümerIn: " ++ p)
print_dat (AbstrDatentraeger id t v a) (Version nr sw) = 
    do putStrLn "Abstrakter Datenträger"
       putStrLn (" ID: " ++ (show id))
       putStrLn (" Typ: " ++ t)
       putStrLn (" Version: " ++ (show nr))
       putStrLn (" Adresse: " ++ a)

-- Überprüfe ob der Datenträger existiert und kein abstrakter Datenträger ist.
dat_test :: [Datentraeger] -> DatentraegerId -> IO()
dat_test _ (-1)               = putStrLn "Error: Ungültige Datenträger ID."
dat_test [] _                 = putStrLn ("Error: Datentraeger nicht gefunden.")
dat_test ((PhysDatentraeger id t v p):dts) id_in
    | id == id_in             = putStr ""
    | otherwise               = dat_test dts id_in
dat_test ((AbstrDatentraeger id t v a):dts) id_in
    | id == id_in             = putStrLn "Error: Abstrakter Datentraeger kann nicht verkauft werden."
    | otherwise               = dat_test dts id_in

-- Verkauft einen Datenträger, das heißt die EigentümerIn des Datenträgers wird geändert (falls dieser existiert)
sell :: [Datentraeger] -> DatentraegerId -> Person -> [Datentraeger]
sell dt (-1) _                = dt
sell [] _ _                   = []
sell ((PhysDatentraeger id t v p):dts) id_in eig
    | id == id_in             = ((PhysDatentraeger id t v eig):dts)
    | otherwise               = (PhysDatentraeger id t v p):(sell dts id_in eig)
sell (dt:dts) id_in eig       = (dt:(sell dts id_in eig))

-- Verleiht einen physikalischen Datenträger an eine Person und fügt die Ausleihe der Liste hinzu.
-- Wenn der Datenträger bereits ausgeliehen ist, wird die Person in der Ausleihe überschrieben.
borrow :: [Ausleihe] -> [Datentraeger] -> DatentraegerId -> Person -> [Ausleihe]
borrow al _ (-1) _            = al
borrow al [] _ _              = al
borrow [] ((PhysDatentraeger id t v p):dts) id_in p_in
    | id == id_in             = [(Ausleihe id_in p_in)]
    | otherwise               = borrow [] dts id_in p_in
borrow [] (dt:dts) id_in p_in = borrow [] dts id_in p_in
borrow ((Ausleihe id p):als) dt id_in p_in
    | id == id_in             = ((Ausleihe id p_in):als)
    | otherwise               = (Ausleihe id p):(borrow als dt id_in p_in)
    

-- Wandle einen String in eine ID um, wenn der String keiner ganzen Zahl entspricht, dann -1
parseID :: String -> DatentraegerId
parseID "" = -1
parseID s
    | (length s) > (length (filter isDigit s)) = -1
    | otherwise                                = read s::Int


                
-- Testdaten
test_software :: [Software]
test_software = [Software "Microsoft Windows XP" "Bill Gates, Steve Balmer" "2001" "proprietaeres Betriebssystem",
                 Software "Ubuntu GNU/Linux" "Mark Shuttleworth et. al." "2004" "freies Derivat von Debian",
                 Software "Adobe Photoshop" "Thomas Knoll, John Knoll" "1987" "Bildbearbeitungsprogramm"]
                 
test_versionen :: [Software] -> [Version]
test_versionen sw = [Version 1 (sw !! 0),
                     Version 2 (sw !! 0),
                     Version 7 (sw !! 1),
                     Version 8 (sw !! 1),
                     Version 10 (sw !! 2)]

test_datentr :: [Version] -> [Datentraeger]
test_datentr vs = [PhysDatentraeger 1 "CD" [vs !! 0] "Klaus Purer",
                   PhysDatentraeger 2 "CD" [vs !! 0] "Alex Ahammer",
                   PhysDatentraeger 3 "CD" [vs !! 0] "Alex Fischer",
                   PhysDatentraeger 4 "CD" [vs !! 2] "Klaus Purer",
                   PhysDatentraeger 5 "DVD" [(vs !! 2),(vs !! 3)] "Klaus Purer",
                   PhysDatentraeger 6 "DVD" [vs !! 4] "Alex Ahammer",
                   PhysDatentraeger 7 "Festplatte" [vs !! 1] "Alex Fischer",
                   AbstrDatentraeger 8 "Web-Adresse" [vs !! 1] "https://www.informatik.tuwien.ac.at/msdnaa/"]
                   
test_db :: Database
test_db = (sw,vs,(test_datentr vs),[]) where
                sw = test_software
                vs = test_versionen sw
